%include {

/** @brief Lemon parser grammar for Turtle.
 *
 * The `lemon' parser generator executable must be in your PATH:
 * https://sqlite.org/src/doc/trunk/doc/lemon.html
 *
 * TTL EBNF: https://www.w3.org/TeamSubmission/turtle/#sec-grammar-grammar
 */

#include "codec.h"

}

%name TTLParse

%stack_overflow {
    log_error ("Stack oveflow in TTL parsing.");
    state->rc = LSUP_MEM_ERR;
}

%parse_failure {
    log_error ("TTL parse error. Cannot continue.");
    state->rc = LSUP_PARSE_ERR;
}

%syntax_error {
    // Fail immediately on first error.
    yy_parse_failed (yypParser);
}

//%stack_size 1024

%token_prefix "T_"
%token_type { char * }
%token_destructor { (void) state; free ($$); }

%default_type           { char * }

%extra_argument         { LSUP_TTLParserState *state }

// Low- to high-priority.
%left PERIOD .
%left SEMICOLON .
%left COMMA .
%left STRING INTEGER DOUBLE DECIMAL BOOLEAN QNAME BNODE_ID IRIREF .
%nonassoc LANGTAG PREFIX .
%nonassoc COLON .

/*
 * Rules.
 */

turtleDoc   ::= statements EOF .
statements  ::= statements statement .
statements  ::= .

statement 	::= prefixID .
statement   ::= base .
statement   ::= triples .
statement   ::= WS .

prefixID    ::= PREFIX(P) WS IRIREF(N) PERIOD . {
                LSUP_nsmap_add (
                    LSUP_graph_namespace (
                        LSUP_graph_iter_graph (state->it)
                    ), P, N
                );

                free (P);
                free (N);
            }

base        ::= BASE WS IRIREF(D) PERIOD . {
                LSUP_term_free (state->base);
                state->base = LSUP_iriref_new (D, NULL);

                free (D);
            }

triples 	::= subject(S) ows predObjList(L) PERIOD . {
                size_t ct = LSUP_spo_list_add_triples (state->it, S, L);
                state->ct += ct;
                state->rc = LSUP_OK;
                log_trace ("Added %lu triples.", ct);

                LSUP_term_free (S);
                LSUP_conn_list_free (L);
            }
triples 	::= subject(S) ows predObjList(L) SEMICOLON PERIOD . [PERIOD] {
                size_t ct = LSUP_spo_list_add_triples (state->it, S, L);
                state->ct += ct;
                state->rc = LSUP_OK;
                log_trace ("Added %lu triples.", ct);

                LSUP_term_free (S);
                LSUP_conn_list_free (L);
            }

%type predObjList       { LSUP_ConnectionList * }
%destructor predObjList { LSUP_conn_list_free ($$); }
predObjList(A) ::= predicate(P) ows objectList(O) . [SEMICOLON] {
                A = LSUP_conn_list_new(LSUP_CONN_OUTBOUND);
                LSUP_conn_list_add (A, P, O);
            }
predObjList(A) ::= predObjList(L) SEMICOLON predicate(P) ows objectList(O) . {
                LSUP_conn_list_add (L, P, O);
                A = L;
            }

%type objectList { LSUP_Term ** }
%destructor objectList {
    log_debug ("Freeing object list.");
    size_t i = 0;
    while ($$[i]) LSUP_term_free ($$[i++]);
    free ($$);
}
objectList(A) ::= objectList(L) COMMA object(O) . {
                A = LSUP_term_list_add (L, O);
            }
objectList(A) ::= object(O) . [IRIREF] {
                A = calloc (sizeof (*A), 2);
                A[0] = O;
            }

%type subject { LSUP_Term * }
%destructor subject { LSUP_term_free ($$); }
subject 	::= resource .
subject 	::= blank .

%type predicate { LSUP_Term * }
%destructor predicate { LSUP_term_free ($$); }
predicate   ::= resource .
predicate(A)::= RDF_TYPE . {
                A = LSUP_iriref_new (
                        "rdf:type",
                        LSUP_graph_namespace (LSUP_graph_iter_graph (state->it)));
                }

%type object { LSUP_Term * }
%destructor object { LSUP_term_free ($$); }
object 	    ::= resource .
object 	    ::= blank .
object 	    ::= literal .

%type literal { LSUP_Term * }
%destructor literal { LSUP_term_free ($$); }
literal(A)  ::= STRING(D) . {
                A = LSUP_term_new (LSUP_TERM_LITERAL, D, NULL);
                log_trace ("Created plain literal: \"%s\"", A->data);
                free (D);
            }
literal(A)  ::= STRING(D) LANGTAG(L) . {
                A = LSUP_term_new (LSUP_TERM_LT_LITERAL, D, L);
                log_trace ("Created LT-literal: \"%s\"@%s", A->data, A->lang);
                free (D);
                free (L);
            }
literal(A)  ::= STRING(D) DTYPE_MARKER resource(M) . {
                A = LSUP_term_new (LSUP_TERM_LITERAL, D, M);
                log_trace (
                        "Created DT-literal: \"%s\"^^%s",
                        A->data, A->datatype);
                free (D);
            }
literal(A)  ::= INTEGER(D) . {
                A = LSUP_term_new (
                    LSUP_TERM_LITERAL, D,
                    LSUP_iriref_new (
                        "xsd:integer", 
                        LSUP_graph_namespace (LSUP_graph_iter_graph (state->it))
                    )
                );

                free (D);
            }
literal(A)  ::= DOUBLE(D) . {
                A = LSUP_term_new (
                    LSUP_TERM_LITERAL, D,
                    LSUP_iriref_new (
                        "xsd:double", 
                        LSUP_graph_namespace (LSUP_graph_iter_graph (state->it))
                    )
                );
                free (D);
            }
literal(A)  ::= DECIMAL(D) . {
                A = LSUP_term_new (
                    LSUP_TERM_LITERAL, D,
                    LSUP_iriref_new (
                        "xsd:decimal", 
                        LSUP_graph_namespace (LSUP_graph_iter_graph (state->it))
                    )
                );
                free (D);
            }
literal(A)  ::= BOOLEAN(D) . {
                A = LSUP_term_new (
                    LSUP_TERM_LITERAL, D,
                    LSUP_iriref_new (
                        "xsd:boolean", 
                        LSUP_graph_namespace (LSUP_graph_iter_graph (state->it))
                    )
                );
                free (D);
            }

%type blank { LSUP_Term * }
%destructor blank { LSUP_term_free ($$); }
blank(A)    ::= BNODE_ID(D) . {
                A = LSUP_term_new (LSUP_TERM_BNODE, D, NULL);
                log_trace ("Created blank node: _:%s", A->data);

                free (D);
            }
blank(A)    ::= LBRACKET RBRACKET . [BNODE_ID] {
                A = LSUP_term_new (LSUP_TERM_BNODE, NULL, NULL);
                log_trace ("Created empty list BN: _:%s", A->data);
            }
blank(A)    ::= LBRACKET predObjList(L) RBRACKET . [BNODE_ID] {
                A = LSUP_term_new (LSUP_TERM_BNODE, NULL, NULL);
                state->ct += LSUP_spo_list_add_triples (state->it, A, L);
                log_trace ("Created list BN: _:%s", A->data);

                LSUP_conn_list_free (L);
            }
blank       ::= collection . [BNODE_ID]
blank(A)    ::= LPAREN RPAREN . [BNODE_ID] {
                A = LSUP_iriref_new (
                    "rdf:nil",
                    LSUP_graph_namespace (LSUP_graph_iter_graph (state->it))
                );
                log_trace ("Created list terminator: %s", A->data);
            }

// "collection" is the subject of the first collection item.
%type collection { LSUP_Term * }
%destructor collection { LSUP_term_free ($$); }
// Collection triples are added here to the graph.
collection(A) ::= LPAREN itemList(L) RPAREN . {
                A = LSUP_bnode_add_collection (state->it, L);
                // Free the item list.
                size_t i = 0;
                while (L[i]) LSUP_term_free (L[i++]);
                free (L);
            }

%type itemList { LSUP_Term ** }
%destructor itemList {
    log_debug ("Freeing item list.");
    size_t i = 0;
    while ($$[i]) LSUP_term_free ($$[i++]);
    free ($$);
}
itemList(A) ::= itemList(L) ows object(O) . { A = LSUP_term_list_add (L, O); }
itemList(A) ::= object(O) . {
                A = calloc (sizeof (*A), 2);
                A[0] = O;
            }

%type resource { LSUP_Term * }
%destructor resource { LSUP_term_free ($$); }
resource(A) ::= IRIREF(D) . {
                LSUP_Term *rel_iri = LSUP_iriref_new (D, NULL);
                free (D);
                if (state->base) {
                    A = LSUP_iriref_absolute (rel_iri, state->base);
                    LSUP_term_free (rel_iri);
                } else {
                    A = rel_iri;
                }
                log_trace ("Created IRI: <%s>", A->data);
            }
resource(A) ::= QNAME(D) . {
                A = LSUP_iriref_new (
                    D, LSUP_graph_namespace (LSUP_graph_iter_graph (state->it))
                );
                log_trace ("Created IRI: %s", A->data);
                free (D);
            }

ows         ::= WS .
ows         ::= .
